// ==============================
// Fichier:			main.cp
// Projet:			toolserver(1)
// Ecrit par:		Paul Guyot (pguyot@kallisys.net)
// 
// Cr le:			1/6/2004
// Tabulation:		4 espaces
// 
// Copyright:		 2004 by Paul Guyot.
// 					Tous droits rservs pour tous pays.
// Licence:			Licence Rflexive Kallisys
// 					http://www.kallisys.org/reflexive/
// ===========
// $Id$
// ===========

#include <Carbon/Carbon.h>
#include <ApplicationServices/ApplicationServices.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>

// Code crateur de ToolServer
#define kCreator	'MPSX'

#define kEventClass	'misc'
#define kEventID	'dosc'

OSStatus GetToolServerAddress( AEAddressDesc* outAddress, bool* outIsRunning );
OSStatus LaunchToolServer( AEAddressDesc* outAddress );
OSStatus DoScript(
			AEAddressDesc* inAddress,
			const char* inScript,
			FSRef* inStdoutFileRef,
			FSRef* inStderrFileRef );
OSErr FSpGetFullPath(
			const FSSpec* spec,
			short* fullPathLength,
			Handle* fullPath );
OSStatus CreateTempFiles(
			FSRef* outStdoutTempFileRef,
			FSRef* outStderrTempFileRef );
OSStatus DumpTempFile( FSRef* inTempFileRef, FILE* inOutputFile );

// -------------------------------------------------------------------------- //
//  * GetToolServerAddress( AEAddressDesc*, bool* )
// -------------------------------------------------------------------------- //
OSStatus
GetToolServerAddress( AEAddressDesc* outAddress, bool* outIsRunning )
{
	OSStatus theErr = noErr;

	*outIsRunning = false;

	ProcessSerialNumber thePSN;	// PSN avec lequel on itre
	thePSN.highLongOfPSN = 0;
	thePSN.lowLongOfPSN = kNoProcess;

	ProcessInfoRec theProcessInfo;	// Infos sur les processus
	theProcessInfo.processInfoLength = sizeof( ProcessInfoRec );
	theProcessInfo.processName = nil;	// Pas besoin du nom
	theProcessInfo.processAppSpec = nil;	// ni du FSSpec

	// Boucle sur les processus.
	while ( ::GetNextProcess( &thePSN ) == noErr )
	{
		if( ::GetProcessInformation( &thePSN, &theProcessInfo ) == noErr )
		{
			if (
				(theProcessInfo.processType == 'APPL') &&  
				(theProcessInfo.processSignature == kCreator))
			{
				theErr = ::AECreateDesc(typeProcessSerialNumber,
					&thePSN,
					sizeof(thePSN),
					outAddress);
				*outIsRunning = true;
				break;
			}
		}
	}
	
	return theErr;
}

// -------------------------------------------------------------------------- //
//  * LaunchToolServer( AEAddressDesc* )
// -------------------------------------------------------------------------- //
OSStatus
LaunchToolServer( AEAddressDesc* outAddress )
{
	int theErr = noErr;

	do {
		// ToolServer est-il dj ouvert?
		bool toolServerIsRunning = false;
		
		theErr = GetToolServerAddress( outAddress, &toolServerIsRunning );
		if (theErr != noErr)
		{
			(void) ::fprintf(
				stderr, "GetToolServerAddress returned an error (%i)\n",
				theErr );
			break;
		}
		
		// If it's not running, start it.
		if (!toolServerIsRunning)
		{
			FSRef appRef;
			theErr = ::LSGetApplicationForInfo(kLSUnknownType,
								kCreator, nil, kLSRolesAll,
								&appRef, nil);
			if (theErr != noErr)
			{
				(void) ::fprintf(
					stderr, "LSGetApplicationForInfo returned an error (%i)\n",
					theErr );
				break;
			}
			
			LSLaunchFSRefSpec launchSpec;
			launchSpec.appRef = &appRef;
			launchSpec.numDocs = 0;
			launchSpec.itemRefs = nil;
			launchSpec.passThruParams = nil;
			launchSpec.launchFlags = kLSLaunchDontAddToRecents
				|   kLSLaunchDontSwitch
				|	kLSLaunchStartClassic
				|   kLSLaunchAndHide
				|	kLSLaunchNoParams;
			launchSpec.asyncRefCon = nil;
			
			theErr = ::LSOpenFromRefSpec( &launchSpec, nil );
			if (theErr != noErr)
			{
				(void) ::fprintf(
					stderr, "LSOpenFromRefSpec returned an error (%i)\n",
					theErr );
				break;
			}

			(void) ::sleep( 3 );

			theErr = GetToolServerAddress( outAddress, &toolServerIsRunning );
			if (theErr != noErr)
			{
				(void) ::fprintf(
					stderr, "GetToolServerAddress returned an error (%i)\n",
					theErr );
				break;
			}
			
			if (!toolServerIsRunning)
			{
				// XXX
				theErr = -1;
				break;
			}
		}
	} while ( false );
	
	return theErr;
}

// -------------------------------------------------------------------------- //
//  * FSpGetFullPath(const FSSpec* spec, short* fullPathLength, Handle* fullPath)
// -------------------------------------------------------------------------- //
// Pomp directement de MoreFiles.
OSErr
FSpGetFullPath(
		const FSSpec* spec,
		short* fullPathLength,
		Handle* fullPath )
{
	OSErr		result;
	OSErr		realResult;
	FSSpec		tempSpec;
	CInfoPBRec	pb;
	
	*fullPathLength = 0;
	*fullPath = NULL;

	/* Default to noErr */
	realResult = result = noErr;
	
	/* work around Nav Services "bug" (it returns invalid FSSpecs with empty names) */
	if ( spec->name[0] == 0 )
	{
		result = FSMakeFSSpec(spec->vRefNum, spec->parID, spec->name, &tempSpec);
	}
	else
	{
		/* Make a copy of the input FSSpec that can be modified */
		BlockMoveData(spec, &tempSpec, sizeof(FSSpec));
	}
	
	if ( result == noErr )
	{
		if ( tempSpec.parID == fsRtParID )
		{
			/* The object is a volume */
			
			/* Add a colon to make it a full pathname */
			++tempSpec.name[0];
			tempSpec.name[tempSpec.name[0]] = ':';
			
			/* We're done */
			result = PtrToHand(&tempSpec.name[1], fullPath, tempSpec.name[0]);
		}
		else
		{
			/* The object isn't a volume */
			
			/* Is the object a file or a directory? */
			pb.dirInfo.ioNamePtr = tempSpec.name;
			pb.dirInfo.ioVRefNum = tempSpec.vRefNum;
			pb.dirInfo.ioDrDirID = tempSpec.parID;
			pb.dirInfo.ioFDirIndex = 0;
			result = PBGetCatInfoSync(&pb);
			// Allow file/directory name at end of path to not exist.
			realResult = result;
			if ( (result == noErr) || (result == fnfErr) )
			{
				/* if the object is a directory, append a colon so full pathname ends with colon */
				if ( (result == noErr) && (pb.hFileInfo.ioFlAttrib & kioFlAttribDirMask) != 0 )
				{
					++tempSpec.name[0];
					tempSpec.name[tempSpec.name[0]] = ':';
				}
				
				/* Put the object name in first */
				result = PtrToHand(&tempSpec.name[1], fullPath, tempSpec.name[0]);
				if ( result == noErr )
				{
					/* Get the ancestor directory names */
					pb.dirInfo.ioNamePtr = tempSpec.name;
					pb.dirInfo.ioVRefNum = tempSpec.vRefNum;
					pb.dirInfo.ioDrParID = tempSpec.parID;
					do	/* loop until we have an error or find the root directory */
					{
						pb.dirInfo.ioFDirIndex = -1;
						pb.dirInfo.ioDrDirID = pb.dirInfo.ioDrParID;
						result = PBGetCatInfoSync(&pb);
						if ( result == noErr )
						{
							/* Append colon to directory name */
							++tempSpec.name[0];
							tempSpec.name[tempSpec.name[0]] = ':';
							
							/* Add directory name to beginning of fullPath */
							(void) Munger(*fullPath, 0, NULL, 0, &tempSpec.name[1], tempSpec.name[0]);
							result = MemError();
						}
					} while ( (result == noErr) && (pb.dirInfo.ioDrDirID != fsRtDirID) );
				}
			}
		}
	}
	
	if ( result == noErr )
	{
		/* Return the length */
		*fullPathLength = GetHandleSize(*fullPath);
		result = realResult;	// return realResult in case it was fnfErr
	}
	else
	{
		/* Dispose of the handle and return NULL and zero length */
		if ( *fullPath != NULL )
		{
			DisposeHandle(*fullPath);
		}
		*fullPath = NULL;
		*fullPathLength = 0;
	}
	
	return ( result );
}

// -------------------------------------------------------------------------- //
//  * CreateTempFiles( FSRef*, FSRef* )
// -------------------------------------------------------------------------- //
OSStatus
CreateTempFiles( FSRef* outStdoutTempFileRef, FSRef* outStderrTempFileRef )
{
	OSStatus theErr = noErr;
	do {
		char* thePath;
		do {
			thePath = ::tempnam(NULL, "ToolServer-stdout");
			
			// Try to create the file.
			int theResult = ::open( thePath, O_CREAT | O_EXCL, 0644 );
			if (theResult == -1)
			{
				if (errno != EEXIST)
				{
					theErr = errno;
					break;
				}
			} else {
				(void) ::close( theResult );
				break;
			}
		} while (true);
		if (theErr != noErr)
		{
			break;
		}
		
		theErr = ::FSPathMakeRef( (const UInt8*) thePath, outStdoutTempFileRef, NULL );
		if (theErr != noErr)
		{
			break;
		}
		
		do {
			thePath = ::tempnam(NULL, "ToolServer-stderr");
			
			// Try to create the file.
			int theResult = ::open( thePath, O_CREAT | O_EXCL, 0644 );
			if (theResult == -1)
			{
				if (errno != EEXIST)
				{
					theErr = errno;
					break;
				}
			} else {
				(void) ::close( theResult );
				break;
			}
		} while (true);
		if (theErr != noErr)
		{
			break;
		}

		theErr = ::FSPathMakeRef( (const UInt8*) thePath, outStderrTempFileRef, NULL );
		if (theErr != noErr)
		{
			break;
		}
	} while (false);
	
	return theErr;
}

// -------------------------------------------------------------------------- //
//  * DoScript( AEAddressDesc*, const char*, FSRef*, FSRef* )
// -------------------------------------------------------------------------- //
OSStatus
DoScript(
		AEAddressDesc* inAddress,
		const char* inScript,
		FSRef* inStdoutFileRef,
		FSRef* inStderrFileRef )
{
	OSStatus theErr = noErr;
	Handle theStdoutFullPath = nil;
	Handle theStderrFullPath = nil;

	do {
		// Get a FSSpec of the stdout.
		FSSpec someFSSpec;
		theErr = ::FSGetCatalogInfo(
						inStdoutFileRef,
						kFSCatInfoNone,
						NULL /* catalogInfo */,
						NULL /* outName */,
						&someFSSpec,
						NULL /* parentRef */);
		if (theErr != noErr)
		{
			break;
		}
		
		// Get the full path for the stdout.
		short theStdoutFullPathSize;
		theErr = ::FSpGetFullPath(
						&someFSSpec,
						&theStdoutFullPathSize,
						&theStdoutFullPath );
		if (theErr != noErr)
		{
			break;
		}
		
		// Same with stderr.
		theErr = ::FSGetCatalogInfo(
						inStderrFileRef,
						kFSCatInfoNone,
						NULL /* catalogInfo */,
						NULL /* outName */,
						&someFSSpec,
						NULL /* parentRef */);
		if (theErr != noErr)
		{
			break;
		}
		
		// Get the full path.
		short theStderrFullPathSize;
		theErr = ::FSpGetFullPath(
						&someFSSpec,
						&theStderrFullPathSize,
						&theStderrFullPath );
		if (theErr != noErr)
		{
			break;
		}
		
		// Create the event.
		AppleEvent theAppleEventToSend = {typeNull, NULL};
		theErr = ::AECreateAppleEvent(
					kEventClass, kEventID,
					inAddress, kAutoGenerateReturnID,
					kAnyTransactionID, &theAppleEventToSend);

		if (theErr != noErr)
		{
			(void) ::fprintf(
				stderr, "AECreateAppleEvent returned an error (%i)\n", theErr );
			break;
		}

		// Add the script to it.
		size_t theScriptLength = ::strlen(inScript);
		// () >> stdout  stderr || echo {Status}
		int theFullScriptLength = theScriptLength + theStdoutFullPathSize
			+ theStderrFullPathSize + 32;
		char* theFullScript = (char*) ::malloc(theFullScriptLength);
		char* theCursor = theFullScript;
		*theCursor++ = '(';
		(void) ::memcpy( theCursor, (const char*) inScript, theScriptLength );
		theCursor += theScriptLength;
		*theCursor++ = ')';
		*theCursor++ = ' ';
		*theCursor++ = '>';
		*theCursor++ = ' ';
		(void) ::memcpy( theCursor, (const char*) *theStdoutFullPath, theStdoutFullPathSize );
		theCursor += theStdoutFullPathSize;
		*theCursor++ = ' ';
		*theCursor++ = '';
		*theCursor++ = ' ';
		(void) ::memcpy( theCursor, (const char*) *theStderrFullPath, theStderrFullPathSize );
		theCursor += theStderrFullPathSize;
		::sprintf( theCursor, " || echo {Status}" );
		
		theErr = ::AEPutParamPtr(
						&theAppleEventToSend,
						keyDirectObject,
						typeCString,
						theFullScript,
						theFullScriptLength);
		if (theErr != noErr)
		{
			(void) ::fprintf(
				stderr, "AEPutParamPtr returned an error (%i)\n", theErr );
			break;
		}		

		AppleEvent theReplyEvent = {typeNull, NULL};
		theErr = ::AESend(&theAppleEventToSend, &theReplyEvent, kAEWaitReply,
						kAENormalPriority, kAEDefaultTimeout,
						NULL, NULL);

		::AEDisposeDesc(&theAppleEventToSend);
		if (theErr != noErr)
		{
			(void) ::fprintf(
				stderr, "AESend returned an error (%i)\n", theErr );
			break;
		}
		
		// Cast the result to a c string.
		Size theStringSize;
		theErr = ::AESizeOfParam(
						&theReplyEvent,
						keyDirectObject,
						NULL,
						&theStringSize);
		theStringSize++;

		// Skip if there is no result.
		if (theErr == errAEDescNotFound)
		{
			theErr = noErr;
		} else {
			if (theErr != noErr)
			{
				(void) ::fprintf(
					stderr, "AESizeOfParam returned an error (%i)\n", theErr );
				break;
			}

			// Allocate the string.
			char* theString = (char*) ::malloc( theStringSize );
			theErr = ::AEGetParamPtr(
							&theReplyEvent,
							keyDirectObject,
							typeCString,
							NULL /* actualType */,
							theString,
							theStringSize,
							NULL /* actualSize */);
			if (theErr != noErr)
			{
				(void) ::fprintf(
					stderr, "AEGetParamPtr returned an error (%i)\n", theErr );
				break;
			}
			
			// Translate end lines.
			int indexString;
			char* cursorString = theString;
			for (indexString = 0; indexString < theStringSize; indexString++)
			{
				if (*cursorString++ == '\r')
				{
					cursorString[-1] = '\n';
				}
			}
			
			// Cast the string to an integer.
			int theResult;
			if (::sscanf( theString, "%i", &theResult ))
			{
				theErr = theResult;
			}
		}

		::AEDisposeDesc(&theReplyEvent);
	} while ( false );
	
	if (theStdoutFullPath != nil)
	{
		DisposeHandle( theStdoutFullPath );
	}
	if (theStderrFullPath != nil)
	{
		DisposeHandle( theStderrFullPath );
	}
	
	return theErr;
}

// -------------------------------------------------------------------------- //
//  * DumpTempFile( FSRef*, FILE* )
// -------------------------------------------------------------------------- //
OSStatus
DumpTempFile(
		FSRef* inTempFileRef,
		FILE* inOutputFile )
{
	OSStatus theErr;
	char* theBuffer = NULL;
	SInt16 theRef = 0;
	
	do {
		HFSUniStr255 theDataForkName;
		theErr = ::FSGetDataForkName( &theDataForkName );
		if (theErr != noErr)
		{
			(void) ::fprintf(
				stderr, "FSGetDataForkName returned an error (%i)\n", theErr );
			break;
		}

		theErr = ::FSOpenFork(
						inTempFileRef,
						theDataForkName.length,
						theDataForkName.unicode,
						fsRdPerm,
						&theRef );
		if (theErr != noErr)
		{
			(void) ::fprintf(
				stderr, "FSOpenFork returned an error (%i)\n", theErr );
			break;
		}

		theBuffer = (char*) ::malloc( 4096 );
		ByteCount theCount = 4096;

		do {
			theErr = ::FSReadFork( theRef, fsAtMark, 0, theCount, theBuffer, &theCount );
			if (theErr == eofErr)
			{
				theErr = noErr;
			}
			
			if (theErr != noErr)
			{
				(void) ::fprintf(
					stderr, "FSReadFork returned an error (%i)\n", theErr );
				break;
			}
			
			// Translate endlines.
			char* bufferCursor = theBuffer;
			char* bufferEnd = &theBuffer[theCount];
			while (bufferCursor != bufferEnd) {
				if (*bufferCursor++ == '\r')
				{
					bufferCursor[-1] = '\n';
				}
			}
			
			theCount = ::fwrite( theBuffer, 1, theCount, inOutputFile );
		} while (theCount == 4096);

		if (theErr != noErr)
		{
			::free( theBuffer );
			break;
		}

	} while ( false );

	(void) ::FSCloseFork( theRef );
	(void) ::FSDeleteObject( inTempFileRef );

	if (theBuffer)
	{
		::free( theBuffer );
	}
	
	return theErr;
}

// -------------------------------------------------------------------------- //
//  * main
// -------------------------------------------------------------------------- //
int
main( int argc, char * const argv[] )
{
	OSStatus theErr = noErr;
	int theResult = 0;

	do {
		AEAddressDesc theTargetDesc;
		theErr = LaunchToolServer(&theTargetDesc);
		if (theErr != noErr)
		{
			(void) ::fprintf(
				stderr, "LaunchToolServer returned an error (%i)\n", theErr );
			break;
		}
		
		// Create the files.
		FSRef theStdout;
		FSRef theStderr;
		theErr = ::CreateTempFiles( &theStdout, &theStderr );
		if (theErr != noErr)
		{
			(void) ::fprintf(
				stderr, "CreateTempFiles returned an error (%i)\n", theErr );
			break;
		}
		
		// Execute the script.
		int argIndex;
		for (argIndex = 1; argIndex < argc; argIndex++)
		{
			theResult = DoScript(&theTargetDesc, argv[argIndex], &theStdout, &theStderr);
			if (theResult != 0)
			{
				break;
			}
		}

		::AEDisposeDesc(&theTargetDesc);
		
		// Catenate the stdout and the stderr to stdout and stderr.
		theErr = ::DumpTempFile( &theStdout, stdout );
		if (theErr != noErr)
		{
			(void) ::fprintf(
				stderr, "DumpTempFile [stdout] returned an error (%i)\n", theErr );
		}
		
		theErr = ::DumpTempFile( &theStderr, stderr );
		if (theErr != noErr)
		{
			(void) ::fprintf(
				stderr, "DumpTempFile [stderr] returned an error (%i)\n", theErr );
		}
	} while( false );
	
	if (theErr != noErr)
	{
		theResult = theErr;
	}
	
	return theResult;
}

// =================================== //
// Your password is pitifully obvious. //
// =================================== //
